using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Text;
using ABC.Tools.Dsl;
using System.CodeDom.Compiler;
using System.CodeDom;
using Microsoft.CSharp;
using System.Reflection;
using System.IO;
using Microsoft.VisualStudio.Modeling;

namespace ABC.Tools.Dsl.CodeGenerator
{
    public class CodeGenerationHelper
    {
        private CodeDomProvider _provider = null;
        private Model _model = null;


        public CodeGenerationHelper(Model model)
        {
            //Perform usual checks
            if (model == null) { throw new ArgumentNullException("model"); }
            
            //if (model.Language == CodeLanguage.CSharp)
            //{
            //    this._provider = new CSharpCodeProvider();
            //}
            //else
            //{
            //    this._provider = new VBCodeProvider();
            //}
            this._model = model;
            this._provider = new CSharpCodeProvider();
        }

        public string Generate()
        {
            CodeCompileUnit unit = new CodeCompileUnit();
            GenerateNamespaces(unit);
            return this.GenerateCode(unit);
        }

        public string GenerateCode(CodeCompileUnit unit)
        {
            //Perform usual checks
            if (unit == null) { throw new ArgumentNullException("unit"); }

            StringBuilder builder = new StringBuilder();
            using (StringWriter writer = new StringWriter(builder))
            {
                this._provider.GenerateCodeFromCompileUnit(unit, writer, new CodeGeneratorOptions());
            }
            return builder.ToString();
        }

        private void GenerateNamespaces(CodeCompileUnit unit)
        {
            //Perform usual checks
            if (unit == null) { throw new ArgumentNullException("unit"); }
            
            foreach(Namespace nameSpace in this._model.Namespaces)
            {
                //Generate and add the Namespace
                CodeNamespace cNameSpace = new CodeNamespace(nameSpace.Name);
                
                if(nameSpace is StatefulNamespace)
                {
                    CodeAttributeDeclaration attr = new CodeAttributeDeclaration();
                    attr.Name = "Core.abc.state.application.StatefulNamespace";
                    string attrident = (nameSpace as StatefulNamespace).Identifier;
                    attr.Arguments.Add(new CodeAttributeArgument(new CodePrimitiveExpression(attrident)));
                    unit.AssemblyCustomAttributes.Add(attr);
                }

                //Generate Classes for the Namespace
                foreach (Class cls in nameSpace.Classes)
                {
                    this.GenerateClass(cls, cNameSpace);
                }
                
                //Generate Interfaces for the Namespace
                foreach (Interface inf in nameSpace.Interfaces)
                {
                    this.GenerateInterface(inf, cNameSpace);
                }

                unit.Namespaces.Add(cNameSpace);
            }       
        }

        private void GenerateClass(Class cls, CodeNamespace cNameSpace)
        {
            //Perform usual checks
            if (cls == null) { throw new ArgumentNullException("cls"); }
            if (cNameSpace == null) { throw new ArgumentNullException("cNameSpace"); }

            //Generate the Class itself
            CodeTypeDeclaration ctdecl = new CodeTypeDeclaration(cls.Name);
            ctdecl.TypeAttributes = TypeAttributes.Public;
            ctdecl.IsPartial = true;
            
            //Set super class
            if (cls.SuperClass != null) ctdecl.BaseTypes.Add(cls.SuperClass.Name);
            
            //Set implementation interfaces
            foreach (Interface inf in cls.Interfaces)
            { ctdecl.BaseTypes.Add(inf.Name); }

            //Add the StatefulClass Attribute if Statefull class
            if (cls is StatefulClass)
            {
                CodeAttributeDeclaration attr = new CodeAttributeDeclaration();
                attr.Name = "Core.abc.state.application.StatefulClass";
                //string attrident = (cls as StatefulClass).Identifier;
                //attr.Arguments.Add(new CodeAttributeArgument(new CodePrimitiveExpression(attrident)));
                ctdecl.CustomAttributes.Add(attr);
            }

            //Add the operations
            foreach(ClassOperation op in cls.Operations)
            {
                CodeMemberMethod method = new CodeMemberMethod();
                method.Name = op.Name;
                method.Attributes = MemberAttributes.Public;
                method.ReturnType = new CodeTypeReference(op.Type);
                //method.Parameters.Add

                ctdecl.Members.Add(method);
            }

            //Add the attributes
            foreach (Attribute att in cls.Attributes)
            {
                GenerateField(att, ctdecl);
            }


            ////Add attributes through associations
            ReadOnlyCollection<ClassHasAssociatedClasses> links
                = ClassHasAssociatedClasses.GetLinksToAssociations(cls);
            foreach (ClassHasAssociatedClasses link in links)
            {
                CodeMemberField field = new CodeMemberField(link.TargetClass.Name, link.Name);
                ctdecl.Members.Add(field);
            }

            

            //Add the class to the namespace
            cNameSpace.Types.Add(ctdecl);
        }

        private void GenerateField(Attribute att, CodeTypeDeclaration ctdecl)
        {
            CodeMemberField field = new CodeMemberField(att.Type, att.Name);
                
            //Add the StatefulComponent Attribute if Statefull Attribute
            if (att is StatefulAttribute)
            {
                CodeAttributeDeclaration attr = new CodeAttributeDeclaration();
                attr.Name = "Core.abc.state.application.StatefulComponent";
                string attrident = (att as StatefulAttribute).Identifier;
                attr.Arguments.Add(new CodeAttributeArgument(new CodePrimitiveExpression(attrident)));
                field.CustomAttributes.Add(attr);
            }

            ctdecl.Members.Add(field);
        }

        private void GenerateInterface(Interface inf, CodeNamespace cNameSpace)
        {
            //Perform usual checks
            if (inf == null) { throw new ArgumentNullException("inf"); }
            if (cNameSpace == null) { throw new ArgumentNullException("cNameSpace"); }

            //Generate the interface itself
            CodeTypeDeclaration ctdecl = new CodeTypeDeclaration(inf.Name);
            ctdecl.TypeAttributes = TypeAttributes.Public;
            ctdecl.IsInterface = true;

            //Add the operations
            foreach (InterfaceOperation op in inf.Operations)
            {
                CodeMemberMethod method = new CodeMemberMethod();
                method.Name = op.Name;
                method.ReturnType = new CodeTypeReference(op.Type);
                ctdecl.Members.Add(method);
            }

            //Add the interface to the namespace
            cNameSpace.Types.Add(ctdecl);
        }
    }
}
